#define ATmega8515
#define DEVICE_ADDRESS 0xFE

//#define F_CPU            16000000

#include <avr/io.h>
#include <avr/crc16.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

unsigned char priority=2;

/********  INCLUDE HERE  *************/
#include "..\bus\bus.c"

#define C_BUFFER_SIZE 40

/* Device specific defines */
#define MAX3110SCK 7
#define MAX3110MOSI 5
#define MAX3110CS 4 //this is SS pin on PortB
#define MAX3110_RXMASK 0x80
#define MAX3110_TXMASK 0x40
#define SHUT_OFF_PKTS  0x20
#define CMD_READY    0x40

const char NibbleToHex[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

unsigned char console_rxbuffer[C_BUFFER_SIZE];
unsigned char console_txbuffer[C_BUFFER_SIZE];
unsigned char console_cmdbuffer[C_BUFFER_SIZE];

unsigned char crx_head=0;
unsigned char crx_tail=0;
unsigned char ctx_head=0;
unsigned char ctx_tail=0;
unsigned char ccmd_ptr=0;

unsigned char c_high;
unsigned char c_low;

void Configure3110UART(void);
void MAX3110_SPIXmit(void);
void MAX3110_PutChar(unsigned char c);
unsigned char HexToBinary(unsigned char h, unsigned char l);
void PktHandler(void);
void CmdHandler(void);
void init(void);
unsigned char MAX3110_GetChar(void);


SIGNAL(SIG_INTERRUPT1)
{
	unsigned char i;
	/* MAX3110 (Console UART) Receive Handler */
	i = 1;
	while(i)
	{
		c_high = 0x00;
		c_low = 0x00;
		MAX3110_SPIXmit();
		/* Receive byte if available */
		if (c_high & MAX3110_RXMASK)
		{
			/* crx_head points to the next byte to be filled */
			console_rxbuffer[crx_head++] = c_low;
			MAX3110_PutChar(c_low);
			if (c_low == ':') state = state | SHUT_OFF_PKTS;
			if (c_low == 0x0D) state = state | CMD_READY;
			if (crx_head >= C_BUFFER_SIZE) crx_head = 0;
		}
		/* TX next byte if available */
		if (c_high & MAX3110_TXMASK)
		{
			if (ctx_tail != ctx_head)
			{
				/* There's something to xmit */
				c_high = 0x80;
				c_low = console_txbuffer[ctx_tail++];
				MAX3110_SPIXmit();
				if (ctx_tail >= C_BUFFER_SIZE) ctx_tail=0;
				if (ctx_tail == ctx_head)
				{
					c_high = 0xC4;
					c_low = 0x01;
					MAX3110_SPIXmit();
				} else {
					c_high = 0xCC;
					c_low = 0x01;
					MAX3110_SPIXmit();
				}
			}
		}
		i = c_high & MAX3110_RXMASK;
	}
}

int main(void)
{
	unsigned char i,t;
	dev_addr = DEVICE_ADDRESS;
	init();
	sei();
	
	while(1)
	{
		if (state & CMD_READY) CmdHandler();

		if (state & RX_PKT_READY) PktHandler();
		while (state & TX_PKT_READY)
		{
			if (state & RX_PKT_READY) PktHandler();
			i = xmit_packet();
			if (i)
			{
				//delay 10 milliseconds
				for(t=0; t<100; t++)
				{
					TCNT0=0x00;
					while(TCNT0 < __100_US)
					{
						if (state & RX_PKT_READY) PktHandler();
					}
				}
				loneliness--;
			} else {
				cli();
				state = state & (~TX_PKT_READY);
				state = state | TX_BUF_ACTIVE;
				SER_CONT_REG = _BV(RXEN) | _BV(TXEN) | _BV(UDRIE);
				sei();
			}
			dev_addr = DEVICE_ADDRESS;
		}
	}
}

void MAX3110_SPIXmit(void)
{
	PORTB &= ~_BV(MAX3110CS);
	SPDR = c_high;
	while(!(SPSR & _BV(SPIF)));
	c_high = SPDR;
	SPDR = c_low;
	while(!(SPSR & _BV(SPIF)));
	c_low = SPDR;
	PORTB |= _BV(MAX3110CS);
	return;
}

void MAX3110_PutChar(unsigned char c)
{
	unsigned char i = ctx_head;
	i++;
	if (i >= C_BUFFER_SIZE) i = 0;

	while (i == ctx_tail);

	/* crx_head points to the next byte to be filled */
	console_txbuffer[ctx_head++] = c;
	if (ctx_head >= C_BUFFER_SIZE) ctx_head = 0;
	PORTD &= ~_BV(3);
	DDRD |= _BV(3);
	DDRD &= ~_BV(3);
	return;
}

unsigned char MAX3110_GetChar()
{
	unsigned char i;
	/* crx_head points to the next byte to be filled */
	if (crx_tail == crx_head) return(0);
	i = console_rxbuffer[crx_tail++];
	if (crx_tail >= C_BUFFER_SIZE) crx_tail = 0;
	return(i);
}

unsigned char HexToBinary(unsigned char h, unsigned char l)
{
	unsigned char i;
	for (i=0; i<16; i++)
	{
		if (h == NibbleToHex[i])
		{
			h = i*16;
			break;
		}
	}
	
	if (16 == i) return(0);

	for (i=0; i<16; i++)
	{
		if (l == NibbleToHex[i])
		{
			h = h + i;
			break;
		}
	}
	
	if (16 == i) return(0);

	return(h);
	
}
void CmdHandler(void)
{
	unsigned char c, i;

	for (c=0; c<C_BUFFER_SIZE; c++)
	{
		console_cmdbuffer[c] = 0;
	}

	c = MAX3110_GetChar();
	ccmd_ptr = 0;
	
	while(c)
	{
		console_cmdbuffer[ccmd_ptr] = c;	
		ccmd_ptr++;
		
		MAX3110_PutChar(c);
		c = MAX3110_GetChar();
	}
	
	if (console_cmdbuffer[0] != ':') goto bad_cmd;
	if (console_cmdbuffer[(ccmd_ptr-2)] != ';') goto bad_cmd;
	tx_buffer[PKT_SRC] = HexToBinary(console_cmdbuffer[1], console_cmdbuffer[2]);	
	if ('-' != console_cmdbuffer[3] || '>' != console_cmdbuffer[4]) goto bad_cmd;
	tx_buffer[PKT_DEST] = HexToBinary(console_cmdbuffer[5], console_cmdbuffer[6]);

	c = 5;
    i = 8;
	while(console_cmdbuffer[i] != 0 && i<C_BUFFER_SIZE)
	{
		tx_buffer[c] = HexToBinary(console_cmdbuffer[i], console_cmdbuffer[i+1]);
		c=c+1;
		i=i+3;
	}

	tx_buffer[PKT_LEN] = c-1;	
	
	state = state | TX_PKT_READY;
	state = state & (~SHUT_OFF_PKTS);

	MAX3110_PutChar('O');
	MAX3110_PutChar('k');
	MAX3110_PutChar(0x0A);
	MAX3110_PutChar(0x0D);
	MAX3110_PutChar(0x0A);
	MAX3110_PutChar(0x0D);

	state = state & (~CMD_READY);

	return;

bad_cmd:

	MAX3110_PutChar('?');
	MAX3110_PutChar('?');
	MAX3110_PutChar('?');
	MAX3110_PutChar(0x0A);
	MAX3110_PutChar(0x0D);
	MAX3110_PutChar(0x0A);
	MAX3110_PutChar(0x0D);

	state = state & (~SHUT_OFF_PKTS);
	state = state & (~CMD_READY);

	return;
}


void PktHandler(void)
{
	unsigned char i, crc_low, crc_high;
	unsigned short crc;
	unsigned char CRC16_High=0, CRC16_Low=0;
	crc = 0x0000;
	/*************** PACKET FILTER ***************/

	/* Loopback Test - did we send it? */
/*	if (rx_buffer[PKT_SRC] == dev_addr) 
	{
		goto	PktIgnore;
	}*/

	/* Destination Test - is this for us? */
/*	if (rx_buffer[PKT_DEST] != 0xFF && rx_buffer[PKT_DEST] != dev_addr) 
	{
		goto	PktIgnore;
	}*/

	/* CRC16 Test - is the packet intact? */
	
	for(i=0; i < rx_buffer[PKT_LEN]; i++)
	{
		if ((i != PKT_CRC_H) && (i != PKT_CRC_L))
		{
			crc = _crc16_update(crc, rx_buffer[i]);
		}
	}
	
	crc_low = crc & 0xFF;
	crc_high = crc >> 8;
	
	if((crc_low != rx_buffer[PKT_CRC_L]) || (crc_high != rx_buffer[PKT_CRC_H])) goto PktError;
	/*************** END PACKET FILTER ***************/

	/*************** PACKET SUCCESS - PROCESS HERE ***************/
	if (state & SHUT_OFF_PKTS)
	{
		state = state & (~RX_PKT_READY);
		return;
	}			
	for (i=0; i<rx_buffer[PKT_LEN]; i++)
	{
		CRC16_High = rx_buffer[i] & 0xF0;
		CRC16_Low = rx_buffer[i] & 0x0F;

		MAX3110_PutChar(NibbleToHex[(CRC16_High>>4)]);
		MAX3110_PutChar(NibbleToHex[(CRC16_Low)]);
		MAX3110_PutChar(' ');
	}
	MAX3110_PutChar(0x0A);
	MAX3110_PutChar(0x0D);


	state = state & (~RX_PKT_READY);
	return;


PktError:

	/*************** PACKET ERROR ***************/

PktIgnore:
	state = 0;
	return;
}

void init(void) //initializes the U(S)ART and the timer
{
	// set baud rate
	UBRRH=0x00;
	UBRR=UART_BAUD_SELECT;
	
	RS485DDR = _BV(RS485TXEN) | _BV(TX); //set TXen pin as output
	RS485PORT = 0x00;	// and low
	
	// Format: 8-N-1
	SER_CONT_REG = _BV(TXEN) | _BV(RXEN) | _BV(RXCIE);
	UCSRC = _BV(URSEL) | _BV(UCSZ1) | _BV(UCSZ0);
	
	// prescaler is 8, timer counts 1/2 microseconds
	TCCR0 = 0x02;	
	/* prescaler is 1024 for 16bit timer */
	TCCR1A = 0x00;
	TCCR1B = 0x00 | _BV(CS12) | _BV(CS10);
	
	DDRC=0xFF;
	PORTC=0x01;
		
	ACSR = _BV(ACD);
	SFIOR=_BV(PUD);
	
	//Initialization of SPI	
	/* Set MOSI, CS and SCK to output, all others to input (CS is the /SS line)*/
	DDRB = _BV(MAX3110SCK)|_BV(MAX3110MOSI)|_BV(MAX3110CS);
	PORTB |= _BV(MAX3110CS);
	
	/* Enable SPI, Master, set clock rate fck/16 */
	// CPOL and CPHA are zero: idle is low, data is sampled on rising edge (?)
	SPCR = _BV(SPE)|_BV(MSTR)|_BV(SPR0);

	//Configuration word for MAX3110:
	//FIFO is enabled, shutdown deactivated, rx interrupt enabled, baud rate is 115200bps
	c_high = 0xCC;
	c_low = 0x01;
	MAX3110_SPIXmit();
	
	MCUCR = _BV(ISC11);
	GICR = _BV(INT1);
}
